﻿using System;
using System.Windows;
using System.Windows.Data;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Controls;
using System.Diagnostics;
using System.Collections;
using System.Windows.Threading;

namespace WpfApplication1
{
    /***********************基本原理*********************************************
     * 假设有几个Button控件要在UniformGrid布局和StackPanel布局之间切换,我们可以这样来实现:
     * 就其中的一个按钮btn1而言:
     * 无论btn1被加载到哪个面板中,都是先从上一个面板中将btn1删除,然后在调用另外一个面板的Children.Add方法,
     * 该动作都是瞬间完成的,除非修改面板内部逻辑,不可以产生动画,更不可能有过度效果.那么事实上,我们既不将btn1放到
     * UnifromGrid中也不把它放到StackPanel中,而是将它放到一个Canvas中.并且Canvas与UnifromGrid以及
     * StackPanel重合. 
     * 而放到UnifromGrid或StackPanel中的是另外一个元素:我们称为LayoutToLayoutTarget(Target),
     * 切换布局的时候,事实上,我们是将Target从一个面板中删除然后在加到另一个面板中,与此同时,我们的btn1参照Target在新
     * 面板中的位置和大小来作为其动画的目标值,然后利用动画来到变化到此值.
     * 当由于其他原因(比如用户改变窗口大小等)导致Target位置或大小发生变化时,我们的btn1也会立即做出相应的变化(而不采用
     * 动画来渐变了).可以看出btn1始终在Canvas中,只是大小和位置发生了变化,而让人产生了错觉
     * 这里的btn1只是一个特例,为了让所有的控件都可以达到这个效果而不仅仅是Button,我们就使用LayoutToLayoutHost
     * 来包含其他控件(host.Child=myBtn).Host的行为为上面的btn1完全一样.
     ************************************************************************/

    /************************************************
     * 关于子控件在父控件中的位置,若父控件是Canvas,则GetTop和GetLeft,
     * 而更通用的办法是child.TranslatePoint(new Point(0, 0), Parent)
     * 它的意思是:子控件上(0,0)点相对与父控件的坐标
     * 当然这也适用于任意两个控件之间
     ************************************************/



    /// <summary>
    /// LayoutToLayoutTarget用来指示LayoutToLayoutHost的大小以及将放置在什么位置.当布局变化时,先将LayoutToLayoutTarget变化到
    /// 合适的位置和大小,然后LayoutToLayoutHost再根据LayoutToLayoutTarget的大小和位置来进行动画
    /// </summary>
    public class LayoutToLayoutTarget : Border
    {
        /*
         * Create a dependency property for binding the host to the target
         * */
        public static readonly DependencyProperty HostProperty =
            DependencyProperty.Register("Host", typeof(LayoutToLayoutHost), typeof(LayoutToLayoutTarget), null);

        /// <summary>
        /// 获取或设置其对应的Host
        /// </summary>
        public LayoutToLayoutHost Host
        {
            get
            {
                return (LayoutToLayoutHost)GetValue(HostProperty);
            }
            set
            {
                SetValue(HostProperty, value);
            }
        }

        /*
         * If the layout changes for the target (the object that actually gets passed between layout schemes)
         * then let the host know so that it can update itself if it wants to (i.e. if it's not in the middle of
         * an animation).
         * */
        public LayoutToLayoutTarget()
        {
            //当Target布局发生变化后,通知Host,以便Host来以动画的形式转换到新布局中
            LayoutUpdated += UpdateHost;
        }

        void UpdateHost(object sender, EventArgs e)
        {
            if (Host != null)
                Host.UpdateFromTarget();
        }
    }

    /// <summary>
    /// LayoutToLayoutHost用来Host控件.LayoutToLayoutTarget用来指示LayoutToLayoutHost的大小以及将放置在什么位置.当布局变化时,先将LayoutToLayoutTarget变化到
    /// 合适的位置和大小,然后LayoutToLayoutHost再根据LayoutToLayoutTarget的大小和位置来进行动画
    /// </summary>
    public class LayoutToLayoutHost : Border
    {
        //Should I animate my way to future layout changes?
        public bool AnimateChanges = false;

        /// <summary>
        /// 指示是否正在动画当中
        /// </summary>
        public bool Animating = false;

        /// <summary>
        /// 动画时间
        /// </summary>
        int timeSpan = 500;

        /// <summary>
        /// 指示是否是淡出动画
        /// </summary>
        private bool DisappearOnCompletion = false;

        /// <summary>
        /// 用于动画的TranslateTransform
        /// </summary>
        private TranslateTransform translation;

        //This keeps track of time for when all the animations finish
        private DispatcherTimer Refresher;

        public static readonly DependencyProperty TargetProperty =
            DependencyProperty.Register("Target", typeof(LayoutToLayoutTarget), typeof(LayoutToLayoutHost), null);

        /// <summary>
        /// 其对应的Traget
        /// </summary>
        public LayoutToLayoutTarget Target
        {
            get
            {
                return (LayoutToLayoutTarget)GetValue(TargetProperty);
            }
            set
            {
                SetValue(TargetProperty, value);
            }
        }


        public LayoutToLayoutHost()
        {
            Loaded += OnLoad;
        }

        void OnLoad(object sender, RoutedEventArgs e)
        {
            if (Target != null)
            {
                Debug.WriteLine("Binding Target!!!");
                BindToTarget(Target);
            }
            else
                Debug.WriteLine("Target was NULL!");

            Unloaded += OnUnload;   //this gets done here rather in the ctor to avoid a frame bug (Windows OS 1224171)
        }

        /*
         * If the host gets unloaded, break the cyclical reference to make sure the GC does its job
         * */
        void OnUnload(object sender, RoutedEventArgs e)
        {
            Target.Host = null;
            Target = null;
        }

      
        /// <summary>
        /// 将Host和Target关联起来
        /// </summary>
        /// <param name="t"></param>
        public void BindToTarget(LayoutToLayoutTarget t)
        {
            Target = t;
            t.Host = this;
            translation = new TranslateTransform(0, 0);
            RenderTransform = translation;

            Refresher = new DispatcherTimer();
            Refresher.Interval = TimeSpan.FromMilliseconds(timeSpan);
            Refresher.Tick += OnAnimStateInvalidated;

            UpdateFromTarget();
        }

        /// <summary>
        /// 开始一个新动画
        /// </summary>
        /// <param name="Disappear">是否淡出</param>
        public void BeginAnimating(bool Disappear)
        {
            if (Animating == true)
            {
                EndAnimations();
            }

            Animating = false;
            AnimateChanges = true;

            DisappearOnCompletion = Disappear;

            if (Visibility != Visibility.Visible)
            {
                Visibility = Visibility.Visible;
                Opacity = 0.0;
            }
        }

        /*
         * Do an immediate update, as long as there is not an animation running
         * */
        public void UpdateFromTarget()
        {
            if ((Target == null) || (Animating == true))
                return;

            if (AnimateChanges)
                AnimateFromTarget();
            else
                MatchLayout();

        }

        /*
         * The double-check might not be necessary anymore, but this fixed a layout infinite loop
         * */
        private void MatchLayout()
        {
            if (Width != Target.ActualWidth)
                Width = Target.ActualWidth;

            if (Height != Target.ActualHeight)
                Height = Target.ActualHeight;

            //返回的是Target上的(0,0)点相对与Parent的坐标
            Point pt = Target.TranslatePoint(new Point(0, 0), Parent as UIElement);

            TranslateTransform t = RenderTransform as TranslateTransform;

            if (Math.Abs(t.X - pt.X) > 1)
                t.X = pt.X;

            if (Math.Abs(t.Y - pt.Y) > 1)
                t.Y = pt.Y;
        }

        /*
         * Make a local animation for each animated property
         * Base the destination on the new layout position of the target
         * */
        private void AnimateFromTarget()
        {
            int time = timeSpan;
            Animating = true;
            AnimateChanges = false;

            Point pt = Target.TranslatePoint(new Point(0, 0), Parent as UIElement);
            TranslateTransform t = RenderTransform as TranslateTransform;

            BeginAnimation(LayoutToLayoutHost.WidthProperty, SetupDoubleAnimation(Width, Target.ActualWidth, time));
            BeginAnimation(LayoutToLayoutHost.HeightProperty, SetupDoubleAnimation(Height, Target.ActualHeight, time));
            translation.BeginAnimation(TranslateTransform.XProperty, SetupDoubleAnimation(t.X, pt.X, time));
            translation.BeginAnimation(TranslateTransform.YProperty, SetupDoubleAnimation(t.Y, pt.Y, time));

            if (DisappearOnCompletion == true)
                BeginAnimation(LayoutToLayoutHost.OpacityProperty, SetupDoubleAnimation(Opacity, 0.0, time));
            else
                BeginAnimation(LayoutToLayoutHost.OpacityProperty, SetupDoubleAnimation(Opacity, 1.0, time));


            Refresher.IsEnabled = false;        //this restarts the timer
            Refresher.IsEnabled = true;
            Refresher.Start();
        }

        /*
         * This gets called by the DispatcherTimer Refresher when it is done
         * */
        void OnAnimStateInvalidated(object sender, EventArgs e)
        {
            Animating = false;
            Refresher.IsEnabled = false;
            Refresher.Stop();

            EndAnimations();

            if (DisappearOnCompletion == true)
                Visibility = Visibility.Hidden;
        }

        /*
         * Explicitly replace all of the local animations will null
         * */
        public void EndAnimations()
        {
            double xBuffer = translation.X;
            double yBuffer = translation.Y;
            double widthBuffer = Width;
            double heightBuffer = Height;
            double opacityBuffer = Opacity;

            BeginAnimation(LayoutToLayoutHost.WidthProperty, null);
            BeginAnimation(LayoutToLayoutHost.HeightProperty, null);
            translation.BeginAnimation(TranslateTransform.XProperty, null);
            translation.BeginAnimation(TranslateTransform.YProperty, null);
            BeginAnimation(LayoutToLayoutHost.OpacityProperty, null);

            translation.X = xBuffer;
            translation.Y = yBuffer;
            Height = heightBuffer;
            Width = widthBuffer;
            Opacity = opacityBuffer;
        }

        /*
         * Helper function to create a DoubleAnimation
         * */
        public DoubleAnimation SetupDoubleAnimation(double From, double To, int time)
        {
            DoubleAnimation myDoubleAnimation = new DoubleAnimation();
            myDoubleAnimation.From = From;
            myDoubleAnimation.To = To;
            myDoubleAnimation.Duration = new Duration(TimeSpan.FromMilliseconds(time));
            myDoubleAnimation.AutoReverse = false;

            return myDoubleAnimation;
        }

        /*
         * Break the cyclical reference
         * */
        public void ReleaseFromTarget()
        {
            Target.Host = null;
            Target = null;
        }
    }
}
